; -*- Mode: asm; tab-width: 8; c-basic-offset: 4 -*-

; ***** BEGIN LICENSE BLOCK *****
; Version: MPL 1.1/GPL 2.0/LGPL 2.1
; 
; The contents of this file are subject to the Mozilla Public License Version 
; 1.1 (the "License"); you may not use this file except in compliance with 
; the License. You may obtain a copy of the License at 
; http://www.mozilla.org/MPL/
; 
; Software distributed under the License is distributed on an "AS IS" basis,
; WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
; for the specific language governing rights and limitations under the
; License.
; 
; The Original Code is an OS/2 implementation of js_CompareAndSwap in assembly.
; 
; The Initial Developer of the Original Code is
; IBM Corporation.
; Portions created by the Initial Developer are Copyright (C) 2001
; the Initial Developer. All Rights Reserved.
; 
; Contributor(s):
; 
; Alternatively, the contents of this file may be used under the terms of
; either the GNU General Public License Version 2 or later (the "GPL"), or
; the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
; in which case the provisions of the GPL or the LGPL are applicable instead
; of those above. If you wish to allow use of your version of this file only
; under the terms of either the GPL or the LGPL, and not to allow others to
; use your version of this file under the terms of the MPL, indicate your
; decision by deleting the provisions above and replace them with the notice
; and other provisions required by the GPL or the LGPL. If you do not delete
; the provisions above, a recipient may use your version of this file under
; the terms of any one of the MPL, the GPL or the LGPL.
; 
; ***** END LICENSE BLOCK *****

        .486P
        .MODEL FLAT, OPTLINK
        .STACK

        .CODE

;;;---------------------------------------------------------------------
;;; int _Optlink js_CompareAndSwap(jsword *w, jsword ov, jsword nv)
;;;---------------------------------------------------------------------
js_CompareAndSwap     PROC OPTLINK EXPORT
        push ebx
        mov ebx, eax
        mov eax, edx
        mov edx, ebx
        lock cmpxchg [ebx], ecx
        sete al
        and eax, 1h
        pop ebx
        ret
js_CompareAndSwap     endp

        END
